package ru.dbotthepony.mc.otm.block.entity.matter

import com.mojang.serialization.Codec
import com.mojang.serialization.codecs.RecordCodecBuilder
import net.minecraft.core.BlockPos
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.AbstractContainerMenu
import net.minecraft.world.item.ItemStack
import net.minecraft.world.level.Level
import net.minecraft.world.level.block.state.BlockState
import net.minecraftforge.common.ForgeConfigSpec
import ru.dbotthepony.mc.otm.block.entity.JobContainer
import ru.dbotthepony.mc.otm.block.entity.JobStatus
import ru.dbotthepony.mc.otm.block.entity.ItemJob
import ru.dbotthepony.mc.otm.block.entity.MachineJobEventLoop
import ru.dbotthepony.mc.otm.block.entity.MatteryWorkerBlockEntity
import ru.dbotthepony.mc.otm.capability.FlowDirection
import ru.dbotthepony.mc.otm.capability.MatteryCapability
import ru.dbotthepony.mc.otm.capability.UpgradeType
import ru.dbotthepony.mc.otm.capability.energy.BlockEnergyStorageImpl
import ru.dbotthepony.mc.otm.capability.energy.ProfiledEnergyStorage
import ru.dbotthepony.mc.otm.capability.energy.WorkerEnergyStorage
import ru.dbotthepony.mc.otm.capability.matter.*
import ru.dbotthepony.mc.otm.config.EnergyBalanceValues
import ru.dbotthepony.mc.otm.container.HandlerFilter
import ru.dbotthepony.mc.otm.container.MatteryContainer
import ru.dbotthepony.mc.otm.container.UpgradeContainer
import ru.dbotthepony.mc.otm.core.math.Decimal
import ru.dbotthepony.mc.otm.core.math.DecimalConfigValue
import ru.dbotthepony.mc.otm.core.math.defineDecimal
import ru.dbotthepony.mc.otm.core.util.WriteOnce
import ru.dbotthepony.mc.otm.data.DecimalCodec
import ru.dbotthepony.mc.otm.data.UUIDCodec
import ru.dbotthepony.mc.otm.data.minRange
import ru.dbotthepony.mc.otm.graph.matter.MatterNode
import ru.dbotthepony.mc.otm.matter.MatterManager
import ru.dbotthepony.mc.otm.menu.matter.MatterReplicatorMenu
import ru.dbotthepony.mc.otm.registry.MBlockEntities
import ru.dbotthepony.mc.otm.registry.MItems
import ru.dbotthepony.mc.otm.registry.MNames
import java.util.*

class MatterReplicatorBlockEntity(p_155229_: BlockPos, p_155230_: BlockState) :
	MatteryWorkerBlockEntity<MatterReplicatorBlockEntity.ReplicatorJob>(MBlockEntities.MATTER_REPLICATOR, p_155229_, p_155230_, ReplicatorJob.CODEC) {

	class ReplicatorJob(
		itemStack: ItemStack,
		val matterPerTick: Decimal,
		val task: UUID,
		var matterValue: Decimal,
		val pattern: Optional<PatternState>,
		val asDust: Boolean,
		ticks: Double,
	) : ItemJob(itemStack, ticks, BASE_CONSUMPTION) {
		companion object {
			val CODEC: Codec<ReplicatorJob> by lazy {
				RecordCodecBuilder.create {
					it.group(
						ItemStack.CODEC.fieldOf("Item").forGetter(ReplicatorJob::itemStack),
						DecimalCodec.minRange(Decimal.ZERO).fieldOf("matterPerTick").forGetter(ReplicatorJob::matterPerTick),
						UUIDCodec.fieldOf("task").forGetter(ReplicatorJob::task),
						DecimalCodec.minRange(Decimal.ZERO).fieldOf("matterValue").forGetter(ReplicatorJob::matterValue),
						PatternState.CODEC.optionalFieldOf("pattern").forGetter(ReplicatorJob::pattern),
						Codec.BOOL.fieldOf("asDust").forGetter(ReplicatorJob::asDust),
						Codec.doubleRange(0.0, Double.MAX_VALUE).fieldOf("Ticks").forGetter(ReplicatorJob::ticks),
					).apply(it, ::ReplicatorJob)
				}
			}
		}
	}

	override val upgrades = UpgradeContainer(this::setChangedLight, 3, UpgradeType.REPLICATOR)
	val energy = ProfiledEnergyStorage(WorkerEnergyStorage(::energyLevelUpdated, upgrades.transform(ENERGY_VALUES)))
	val matter = ProfiledMatterStorage(MatterStorageImpl(::matterLevelUpdated, FlowDirection.INPUT, upgrades.matterCapacity(::MATTER_CAPACITY)))
	val container = MatteryContainer(::itemContainerUpdated, 5).also(::addDroppableContainer)

	val energyConfig = ConfigurableEnergy(energy)
	val itemConfig = ConfigurableItemHandler(output = container.handler(HandlerFilter.OnlyOut))

	val matterNode = object : MatterNode() {
		override fun getMatterHandler(): IMatterStorage {
			return matter
		}

		override fun onMatterTaskCreated(task: ReplicationTask) {
			jobEventLoops[0].notify(MachineJobEventLoop.IdleReason.OBSERVING)
		}

		override fun onMatterTaskUpdated(newState: ReplicationTask, oldState: ReplicationTask) {
			jobEventLoops[0].notify(MachineJobEventLoop.IdleReason.OBSERVING)
		}

		override fun onPatternAdded(state: PatternState) {
			jobEventLoops[0].notify(MachineJobEventLoop.IdleReason.OBSERVING)
		}
	}

	init {
		exposeGlobally(MatteryCapability.MATTER, matter)
		exposeGlobally(MatteryCapability.MATTER_NODE, matterNode)

		savetables.stateful(::energy, ENERGY_KEY)
		savetables.stateful(::matter, MATTER_STORAGE_KEY)
		savetables.stateful(::container, INVENTORY_KEY)
		savetables.stateful(::upgrades)
	}

	override fun createMenu(containerID: Int, inventory: Inventory, ply: Player): AbstractContainerMenu {
		return MatterReplicatorMenu(containerID, inventory, this)
	}

	override fun onJobFinish(status: JobStatus<ReplicatorJob>, id: Int) {
		val job = status.job

		if (job.asDust) {
			job.matterValue = MItems.MATTER_DUST.moveIntoContainer(job.matterValue, container, OUTPUT_DUST_MAIN, OUTPUT_DUST_STACKING)

			if (!job.matterValue.isZero) {
				return status.throttle()
			}

			matterNode.graph.notifyTaskCompletion(job.task)
		} else {
			if (!container.fullyAddItem(job.itemStack, FIRST_ACTUAL_OUTPUT_SLOT .. LAST_ACTUAL_OUTPUT_SLOT)) {
				return status.noItem()
			}

			matterNode.graph.notifyTaskCompletion(job.task)
		}
	}

	override fun setRemoved() {
		super.setRemoved()
		matterNode.isValid = false
	}

	override fun setLevel(level: Level) {
		super.setLevel(level)
		matterNode.discover(this)
	}

	override fun jobUpdated(new: ReplicatorJob?, old: ReplicatorJob?, id: Int) {
		visualItemStack = new?.itemStack ?: ItemStack.EMPTY
		visualProgress = 0f
	}

	var visualItemStack by synchronizer.item(observe = false)
		private set

	var visualProgress by synchronizer.float().property
		private set

	var renderRotation = 0f
	var lastRender = 0L
	var particleRenderScore = 0L

	override fun tick() {
		super.tick()

		if (jobEventLoops[0].currentJob == null && matter.storedMatter.isPositive) {
			matter.extractMatter(matterNode.graph.receiveMatter(matter.storedMatter, false), false)
		}
	}

	override fun computeNextJob(id: Int): JobContainer<ReplicatorJob> {
		if (energy.batteryLevel < BASE_CONSUMPTION) {
			return JobContainer.noEnergy()
		}

		val allocation = matterNode.graph.allocateTask(simulate = false) ?: return JobContainer.observe()
		val stack = allocation.task.stack(1)
		val matter = MatterManager.get(stack)

		// ????????
		if (!matter.hasMatterValue) return JobContainer.failure()

		val ticks = matter.complexity

		return JobContainer.success(ReplicatorJob(
			itemStack = stack,
			matterPerTick = matter.matter / ticks,
			task = allocation.task.id,
			matterValue = matter.matter,
			pattern = Optional.ofNullable(allocation.pattern),
			asDust = (level?.random?.nextDouble() ?: 1.0) * upgrades.failureMultiplier > (allocation.pattern?.researchPercent ?: 2.0),
			ticks = ticks,
		))
	}

	override fun onJobTick(status: JobStatus<ReplicatorJob>, id: Int) {
		val job = status.job
		val drainPerTick = job.matterPerTick * status.ticksAdvanced

		if (matter.extractMatter(drainPerTick, true) < drainPerTick) {
			val toDrain = (drainPerTick * DRAIN_MULT)
				.coerceAtMost(job.matterPerTick * (status.ticks - status.workTicks + status.ticksAdvanced))
				.coerceAtLeast(Decimal.ONE)
				.coerceAtMost(matter.missingMatter)

			matter.receiveMatter(matterNode.graph.extractMatter(toDrain, false), false)
		}

		status.scale(matter.extractMatter(drainPerTick, false) / drainPerTick)
		visualProgress = status.workProgress
	}

	companion object {
		private val BASE_CONSUMPTION get() = _BASE_CONSUMPTION.get()
		private val DRAIN_MULT get() = _DRAIN_MULT.get()
		private val MATTER_CAPACITY get() = _MATTER_CAPACITY.get()

		private var _BASE_CONSUMPTION: DecimalConfigValue by WriteOnce()
		private var _DRAIN_MULT: DecimalConfigValue by WriteOnce()
		private var _MATTER_CAPACITY: DecimalConfigValue by WriteOnce()

		var ENERGY_VALUES: EnergyBalanceValues by WriteOnce()
			private set

		fun registerConfig(builder: ForgeConfigSpec.Builder) {
			builder.push(MNames.MATTER_REPLICATOR)

			ENERGY_VALUES = BlockEnergyStorageImpl.makeConfigEntry(builder, capacity = Decimal(200_000), throughput = Decimal(4_000))

			_BASE_CONSUMPTION = builder.defineDecimal("basePowerConsumption", Decimal(400), Decimal.ONE)
			_DRAIN_MULT = builder.comment("How much 'ticks' of replication should replicator drain matter from network when running low on internal matter buffer. This is ultimately a performance value.").defineDecimal("drainMultiplier", Decimal(200), Decimal.ONE)
			_MATTER_CAPACITY = builder.defineDecimal("matterCapacity", Decimal(1_000), Decimal.ONE_TENTH)

			builder.pop()
		}

		const val FIRST_ACTUAL_OUTPUT_SLOT = 0
		const val LAST_ACTUAL_OUTPUT_SLOT = 2
		const val OUTPUT_DUST_MAIN = 3
		const val OUTPUT_DUST_STACKING = 4
	}
}
